﻿////////////////////////////////////////////////////////////////////////////////
// Program.cs -- Code to list .NET framework exceptions
//
// Copyright (C) 2009-2011 William Garrison
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option) any
// later version.  This program is distributed in the hope that it will be
// useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
// Public License for more details.
////////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.Xml;
using System.Linq;

using Microsoft.Win32;
using System.IO;

namespace ConsoleApplication1
{
   class Program
   {
      StringBuilder result = new StringBuilder();

      [STAThread] // Some assemblies require this or they fail to load
      static void Main(string[] args)
      {
         new Program().ReportExceptions();
      }

      /// <summary>
      /// Returns the paths of all the assemblies in the .NET Framework
      /// </summary>
      public string[] FindSDKDLLs()
      {
         List<string> allAssemblies = new List<string>();

         RegistryKey DotNetFramework = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Microsoft\.NETFramework\");
         RegistryKey AssemblyFoldersKey = DotNetFramework.OpenSubKey("AssemblyFolders");

         // Current framework path
         string coreSDKPath = (string)DotNetFramework.GetValue("InstallRoot");
         string version = System.Environment.Version.ToString(3);
         string currentSDKPath = System.IO.Path.Combine(coreSDKPath, "v" + version);
         string[] coreSDKAssemblyPaths = System.IO.Directory.GetFiles(currentSDKPath);
         allAssemblies.AddRange(coreSDKAssemblyPaths);
         
         // Additional assemblies in the SDK paths
         foreach (string subkey in AssemblyFoldersKey.GetSubKeyNames())
         {
            // Get the path
            RegistryKey subFolderKey = AssemblyFoldersKey.OpenSubKey(subkey);
            string assemblyPath = (string)subFolderKey.GetValue(null);

            // The directories may not actually exist... lame!
            if (System.IO.Directory.Exists(assemblyPath))
            {
               // Get the DLLs in the path
               string[] assemblyPaths = System.IO.Directory.GetFiles(assemblyPath, "*.dll", System.IO.SearchOption.AllDirectories);

               // Add them to the list
               allAssemblies.AddRange(assemblyPaths);
            }
         }

         // TODO: Also search HKEY_CURRENT_USER... Why would an SDK ever do that?  Well, apparently they do...

         return allAssemblies.ToArray();
      }

      /// <summary>
      /// Produce a list of all exceptions
      /// </summary>
      public void ReportExceptions()
      {
         // Get all the DLLs in the SDK
         string[] dllFiles = FindSDKDLLs();

         // Here's other ways I did this, that are not correct
         // AssemblyName[] references = Assembly.GetExecutingAssembly().GetReferencedAssemblies();
         // string[] assemblies = System.IO.Directory.GetFiles(@"C:\Windows\Microsoft.NET\", "*.dll", System.IO.SearchOption.AllDirectories);
         // foreach (string refName in GAC) // <-- Run gacutil /l to get this

         // Process each one
         foreach (string asmPath in dllFiles)
         {
            Assembly refAsm = null;
            try
            {
               // Load each referenced assembly
               // - Some assemblies don't work with ReflectionOnlyLoad. :(
               //   Assembly refAsm = Assembly.ReflectionOnlyLoad(refName.FullName);
               refAsm = Assembly.LoadFrom(asmPath);

               // Go through all the types in it
               foreach (Type t in refAsm.GetExportedTypes().OrderBy(t => t.Name))
               {
                  // Publically creatable exception?
                  if (t.IsPublic && DerivesFrom(t, typeof(Exception)) && HasPublicConstructor(t))
                  {
                     // Pass asmPath in here because Assembly.Location does not return
                     // always the same path!  Very strange!
                     Report(t, asmPath);
                  }
               }

            }
            catch (Exception)
            {
               // Wow, lots of things go wrong here...
               // swallow 'em, nothing we can do
            }
         }

         // Copy the result to the clipboard for convenience
         System.Windows.Forms.Clipboard.Clear();
         System.Windows.Forms.Clipboard.SetText(result.ToString());
      }

      // Keep track of the last assembly we displayed
      XmlDocument lastXml = null;
      string lastAssembly = null;

      /// <summary>
      /// Load the documentation XML file for a particular assembly
      /// </summary>
      /// <param name="t">The type to load the XML docs for</param>
      /// <param name="originalPath">The original path to the DLL, not the one returned by t.Assembly.Location</param>
      void LoadDocXML(Type t, string originalPath)
      {
         lastXml = new XmlDocument();
         //string dllPath = t.Assembly.Location;
         string dllPath = originalPath;
         string xmlPath = Path.ChangeExtension(dllPath, ".xml");

         // Don't always have 'em :(
         if (File.Exists(xmlPath))
         {
            lastXml.Load(xmlPath);
         }
         else
         {
            xmlPath =
               Path.Combine(@"C:\Program Files\Reference Assemblies\Microsoft\Framework\.NETFramework\v4.0",
                            Path.GetFileNameWithoutExtension(dllPath) + ".xml");
            if (File.Exists(xmlPath))
            {
               lastXml.Load(xmlPath);
            }
         }
      }

      /// <summary>
      /// Get the summary for a particular type, assuming the proper XML is loaded
      /// </summary>
      string GetDescription(Type t)
      {
         string summaryText = ""; // Return blank, not null - saves the caller grief

         // Cut-off at the bracket.  That indicates a type parameter which is not in the XML doc comment name
         string typeName = new string(t.ToString().TakeWhile(c => c != '[').ToArray());

         XmlNode summaryNode = lastXml.SelectSingleNode("doc/members/member[@name='T:" + typeName + "']/summary");
         if (summaryNode != null)
            summaryText = summaryNode.InnerXml;

         return summaryText;
      }

      /// <summary>
      /// Output info for the type
      /// </summary>
      void Report(Type t, string originalPath)
      {
         // Different assembly?  
         if (t.Assembly.FullName != lastAssembly)
         {
            // Then start a new table
            if (lastAssembly != null)
            {
               Output("</table>");
            }

            lastAssembly = t.Assembly.FullName;
            Output("<h1>" + t.Assembly.GetName().Name + "</h1>");
            Output("<table>");

            // Load the XML docs, if available
            LoadDocXML(t, originalPath);
         }

         // Get the description from the doc XML
         string description = GetDescription(t);

         // Ignore deprecated exceptions
         // - For some reason, DirectX has a bunch where the summary says deprecated,
         //   but the class is not marked with ObsoleteAttribute :(
         if (!description.ToLowerInvariant().Contains("deprecated"))
         {
            Output("<tr><td>" + t.ToString() + "</td><td>" + GetDescription(t) + "</td></tr>");
         }
      }

      void Output(string s)
      {
         Console.WriteLine(s);
         result.AppendLine(s);
      }

      /// <summary>
      /// Returns true if the specified type has a public constructor
      /// </summary>
      bool HasPublicConstructor(Type t)
      {
         ConstructorInfo[] constructors = t.GetConstructors();
         foreach (ConstructorInfo constructor in constructors)
         {
            if (constructor.IsPublic)
               return true;
         }
         return false;
      }

      /// <summary>
      /// Returns true if the class dervies from some other class, at some level
      /// </summary>
      bool DerivesFrom(Type t, Type b)
      {
         while (t.BaseType != null)
         {
            // If t is of type b, then we found it
            if (t.BaseType == b)
               return true;

            // Otherwise, go up one level
            t = t.BaseType;
         }
         return false;
      }
   }
}
